import numpy as np
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.messagebox as mb

from . import moduleFrame
from .table import Table, ButtonFrame
from .scrolledFrame import ScrolledFrame


class SpeciationDimerisation(moduleFrame.Strategy):
    def __init__(self, titration):
        self.titration = titration
        titration.freeNames = np.array(["Host"])
        titration.stoichiometries = np.array([[2]])
        titration.boundNames = np.array(["H2"])

    def __call__(self, ks, totalConcs, alphas):
        K = ks[0]
        Htot = totalConcs.T[0]
        H = (-1 + np.sqrt(1 + 8 * Htot * K)) / (4 * K)
        H2 = (1 + 4 * Htot * K - np.sqrt(1 + 8 * Htot * K)) / (8 * K)

        free = np.atleast_2d(H).T
        bound = np.atleast_2d(H2).T
        return free, bound


class SpeciationHG(moduleFrame.Strategy):
    def __init__(self, titration):
        self.titration = titration
        titration.freeNames = np.array(["Host", "Guest"])
        titration.stoichiometries = np.array([[1, 1]])
        titration.boundNames = np.array(["HG"])

    def __call__(self, ks, totalConcs, alphas):
        K = ks[0]
        Htot, Gtot = totalConcs.T
        H = (
            np.sqrt(
                Gtot ** 2 * K ** 2 - 2 * Gtot * K * (Htot * K - 1) + (Htot * K + 1) ** 2
            )
            - Gtot * K
            + Htot * K
            - 1
        ) / (2 * K)
        G = (
            np.sqrt(
                Htot ** 2 * K ** 2 - 2 * Htot * K * (Gtot * K - 1) + (Gtot * K + 1) ** 2
            )
            - Htot * K
            + Gtot * K
            - 1
        ) / (2 * K)
        HG = H * G * K

        free = np.vstack((H, G)).T
        bound = np.atleast_2d(HG).T
        return free, bound


class SpeciationTable(Table):
    def __init__(self, master, titration):
        if hasattr(titration, "freeNames"):
            freeNames = titration.freeNames
        else:
            freeNames = np.array(["Host", "Guest"])
        if hasattr(titration, "boundNames"):
            boundNames = titration.boundNames
        else:
            boundNames = np.array(["HG"])
        if hasattr(titration, "stoichiometries"):
            stoichiometries = titration.stoichiometries
        else:
            stoichiometries = np.array([[1, 1]])

        super().__init__(
            master,
            0,
            0,
            freeNames,
            allowBlanks=False,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("titles", "new", "delete"),
            boldTitles=True,
        )
        for boundName, stoichiometry in zip(boundNames, stoichiometries):
            self.addRow(boundName, stoichiometry)

    def convertData(self, number):
        return int(number)


class SpeciationPopup(tk.Toplevel):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter speciation table")

        height = int(self.master.winfo_height() * 0.4)
        frame = ScrolledFrame(self, height=height, max_width=1500)
        frame.pack(expand=True, fill="both")

        self.innerFrame = frame.display_widget(ttk.Frame, stretch=True)

        self.speciationTable = SpeciationTable(self.innerFrame, titration)
        self.speciationTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(
            self.innerFrame, self.reset, self.saveData, self.destroy
        )
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        self.speciationTable.destroy()
        self.speciationTable = SpeciationTable(self.innerFrame, self.titration)
        self.speciationTable.pack(expand=True, fill="both")

    def saveData(self):
        try:
            stoichiometries = self.speciationTable.data
        except Exception as e:
            mb.showerror(title="Could not save data", message=e, parent=self)
            return

        self.titration.stoichiometries = stoichiometries
        self.titration.freeNames = self.speciationTable.columnTitles
        self.titration.boundNames = self.speciationTable.rowTitles

        self.destroy()


class SpeciationCOGS(moduleFrame.Strategy):
    # TODO: see if this can be sped up using numba
    def COGS(self, M, y, k, alphas):
        free = y.copy()
        bound = np.empty(len(k))
        polymers = np.any(M < 0, 1)
        # the index of the species making up each of the polymers
        polymerParents = np.nonzero(M[polymers])[1]
        complexes = ~polymers

        P = np.empty(len(k))
        P[complexes] = np.sum(M[complexes, :], 1)
        P[polymers] = 2 * k[polymers] * y[polymerParents]
        P = 1 / max(P)
        tol = 1e-7
        while True:
            bound[complexes] = k[complexes] * np.prod(free ** M[complexes, :], 1)
            # cap the maximum guess to avoid divergence
            bound[polymers] = np.where(
                free[polymerParents] * k[polymers] >= 1,
                free[polymerParents] ** 2 * k[polymers] / P,
                (2 - k[polymers] * free[polymerParents])
                * (k[polymers] * free[polymerParents] ** 2)
                / (alphas * (1 - k[polymers] * free[polymerParents]) ** 2),
            )
            total = free + abs(M.T) @ bound  # total concentrations of species
            if all((total - y) <= tol * y):
                break
            # to handle 0 total concentration
            invTotal = np.where(total == 0, 1, 1 / total)
            free *= (y * invTotal) ** P
        return free, bound

    def __call__(self, ks, totalConcs, alphas):
        numPoints = totalConcs.shape[0]
        free = np.zeros((numPoints, self.titration.stoichiometries.shape[1]))
        bound = np.zeros((numPoints, self.titration.stoichiometries.shape[0]))
        for i in range(numPoints):
            free[i], bound[i] = self.COGS(
                self.titration.stoichiometries, totalConcs[i], ks, alphas
            )
        return free, bound


class SpeciationHG2(SpeciationCOGS):
    def __init__(self, titration):
        self.titration = titration
        titration.freeNames = np.array(["Host", "Guest"])
        titration.stoichiometries = np.array([[1, 1], [1, 2]])
        titration.boundNames = np.array(["HG", "HG2"])


class SpeciationCustom(SpeciationCOGS):
    popup = SpeciationPopup
    popupAttributes = ("stoichiometries", "freeNames", "boundNames")


class ModuleFrame(moduleFrame.ModuleFrame):
    frameLabel = "Speciation"
    dropdownLabelText = "Select a binding isotherm:"
    dropdownOptions = {
        "1:1 binding": SpeciationHG,
        "1:2 binding": SpeciationHG2,
        "Dimerisation": SpeciationDimerisation,
        "Custom": SpeciationCustom,
    }
    attributeName = "speciation"



import tkinter as tk
import tkinter.ttk as ttk

import matplotlib
from matplotlib.backends._backend_tk import (
    NavigationToolbar2Tk,
    FigureCanvasTk,
    ToolTip,
    cursord,
)
from matplotlib.backend_bases import _Mode

matplotlib.use("TkAgg")


# updates the navigation toolbar to use ttk, and the correct ttkbootstrap style
def _Button(self, text, image_file, toggle, command):
    if tk.TkVersion >= 8.6:
        PhotoImage = tk.PhotoImage
    else:
        from PIL.ImageTk import PhotoImage
    image = PhotoImage(master=self, file=image_file) if image_file is not None else None
    if not toggle:
        b = ttk.Button(
            master=self,
            text=text,
            image=image,
            command=command,
            style="Outline.TButton",
        )
    else:
        var = tk.IntVar(master=self)
        b = ttk.Checkbutton(
            master=self,
            text=text,
            image=image,
            command=command,
            variable=var,
            style="Outline.Toolbutton",
        )
        b.var = var
    b._ntimage = image
    b.pack(side=tk.LEFT)
    return b


def _update_buttons_checked(self):
    # sync button checkstates to match active mode
    for text, mode in [("Zoom", _Mode.ZOOM), ("Pan", _Mode.PAN)]:
        if text in self._buttons:
            if self.mode == mode:
                self._buttons[text].var.set(1)
            else:
                self._buttons[text].var.set(0)


# make the tooltips appear further to the right to prevent them from
# overlapping the buttons
def showtip(self, text):
    """Display text in tooltip window."""
    self.text = text
    if self.tipwindow or not self.text:
        return
    x, y, _, _ = self.widget.bbox("insert")
    x = x + self.widget.winfo_rootx() + self.widget.winfo_width()
    y = y + self.widget.winfo_rooty()
    self.tipwindow = tw = tk.Toplevel(self.widget)
    tw.wm_overrideredirect(1)
    tw.wm_geometry("+%d+%d" % (x, y))
    try:
        # For Mac OS
        tw.tk.call(
            "::tk::unsupported::MacWindowStyle", "style", tw._w, "help", "noActivates"
        )
    except tk.TclError:
        pass
    label = tk.Label(
        tw, text=self.text, justify=tk.LEFT, relief=tk.SOLID, borderwidth=1
    )
    label.pack(ipadx=1)


def set_cursor(self, cursor):
    try:
        self._tkcanvas.configure(cursor=cursord[cursor])
    except tk.TclError:
        pass


def applyPatch():
    # makes buttons use ttk widgets
    NavigationToolbar2Tk._Button = _Button
    NavigationToolbar2Tk._update_buttons_checked = _update_buttons_checked

    # implements mpl GH PR #22294
    FigureCanvasTk.set_cursor = set_cursor

    # implements mpl GH PR #22078
    ToolTip.showtip = showtip



import os
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.filedialog as fd
import packaging.version

import numpy as np
import tksheet
from scipy.interpolate import interp1d
from cycler import cycler
from ttkbootstrap.widgets import InteractiveNotebook
import tkinter.messagebox as mb
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk, FigureCanvasTkAgg
from matplotlib.figure import Figure

from . import __version__
from . import speciation
from . import equilibriumConstants
from . import totalConcentrations
from . import contributors
from . import proportionality
from . import knownSignals
from . import fitSignals
from . import combineResiduals
from .style import padding
from .scrolledFrame import ScrolledFrame
from .table import Table

warningIcon = "::tk::icons::warning"

titrationModules = [
    speciation,
    equilibriumConstants,
    totalConcentrations,
    contributors,
    proportionality,
    knownSignals,
    fitSignals,
    combineResiduals,
]


class SaveLoadFrame(ttk.Frame):
    def __init__(self, parent, titration, moduleFrames, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.moduleFrames = moduleFrames

        loadOptionsButton = ttk.Button(
            self,
            text="Load Options",
            style="primary.Outline.TButton",
            command=self.loadOptions,
        )
        loadOptionsButton.grid(sticky="nesw", pady=(padding, padding / 2))

        saveOptionsButton = ttk.Button(
            self,
            text="Save Options",
            style="success.Outline.TButton",
            command=self.saveOptions,
        )
        saveOptionsButton.grid(sticky="nesw", pady=padding / 2)

        saveDataButton = ttk.Button(
            self,
            text="Save Processed Data",
            style="success.Outline.TButton",
            command=self.saveData,
        )
        saveDataButton.grid(sticky="nesw", pady=padding / 2)

        self.columnconfigure(0, weight=1)

    def loadOptions(self):
        fileName = fd.askopenfilename(
            title="Load options from file",
            filetypes=[("NumPy archive", "*.npz"), ("all files", "*.*")],
        )
        if fileName == "":
            return

        with np.load(fileName) as options:
            # TODO: check file version against current
            # TODO: show popup to choose which modules should be loaded
            for moduleName, dropdownValue in zip(
                options["moduleNames"], options["dropdownValues"]
            ):
                if moduleName not in self.moduleFrames:
                    continue
                moduleFrame = self.moduleFrames[moduleName]
                if dropdownValue not in moduleFrame.dropdownOptions:
                    continue
                moduleFrame.stringVar.set(dropdownValue)
                attributeName = moduleFrame.attributeName
                Strategy = moduleFrame.dropdownOptions[dropdownValue]
                strategy = Strategy(self.titration)
                setattr(self.titration, attributeName, strategy)
                for attr in strategy.popupAttributes:
                    if attr not in options:
                        continue
                    attrValue = options[attr]
                    if attrValue.shape == ():
                        attrValue = attrValue.item()
                    setattr(self.titration, attr, attrValue)

    def saveOptions(self):
        # TODO: handle errors
        popup = tk.Toplevel()
        popup.title("Choose which options to save")
        popupFrame = ttk.Frame(popup, padding=15)
        popupFrame.pack(expand=True, fill="both")

        label = ttk.Label(popupFrame, text="Choose which options to save:")
        label.pack(pady=(5, 15))

        popup.checkbuttons = {}
        popup.saved = False

        for name, moduleFrame in self.moduleFrames.items():
            checkbutton = ttk.Checkbutton(popupFrame, text=moduleFrame.frameLabel)
            if moduleFrame.stringVar.get() == "":
                checkbutton.state(["disabled"])
            else:
                checkbutton.state(["selected"])
            checkbutton.pack(anchor="w", pady=2.5)
            popup.checkbuttons[name] = checkbutton

        buttonsFrame = ttk.Frame(popupFrame)
        buttonsFrame.pack(expand=True, fill="x")

        cancelButton = ttk.Button(
            buttonsFrame,
            text="Cancel",
            style="secondary.TButton",
            command=popup.destroy,
        )
        cancelButton.grid(row=0, column=0, pady=5)

        saveButton = ttk.Button(
            buttonsFrame,
            text="Save",
            style="success.TButton",
            command=lambda: self.getSelectedOptions(popup),
        )
        saveButton.grid(row=0, column=1, pady=5)

        for col in range(2):
            buttonsFrame.columnconfigure(col, weight=1)

        # Only create the warning label now, so that we know what width to make it
        if (
            (
                isinstance(self.titration.rowFilter, slice)
                and self.titration.rowFilter != slice(None)
            )
            or (
                isinstance(self.titration.rowFilter, np.ndarray)
                and not self.titration.rowFilter.all()
            )
            or (
                type(self.titration.columnFilter) == slice
                and self.titration.columnFilter != slice(None)
            )
            or (
                isinstance(self.titration.columnFilter, np.ndarray)
                and not self.titration.columnFilter.all()
            )
        ):
            # row or column filter is active
            warningLabel = ttk.Label(
                popupFrame,
                text=(
                    "The input data has been modified, so it may need to be saved as a"
                    " universal CSV file for some options to be applied to it."
                ),
            )
            if warningIcon in popup.image_names():
                warningLabel.configure(
                    image=warningIcon,
                    compound="left",
                    wraplength=popup.winfo_reqwidth()
                    - popup.tk.call("image", "width", warningIcon),
                )
            else:
                warningLabel.configure(wraplength=popup.winfo_reqwidth())
            warningLabel.pack(before=label)

        popup.resizable(False, False)
        popup.wait_window()
        if not popup.saved:
            return

        options = {}
        moduleNames = np.array(popup.selectedOptions)

        dropdownValues = np.array([])

        for name in moduleNames:
            moduleFrame = self.moduleFrames[name]
            dropdownValue = moduleFrame.stringVar.get()
            dropdownValues = np.append(dropdownValues, dropdownValue)

            Strategy = moduleFrame.dropdownOptions[dropdownValue]
            for attr in Strategy.popupAttributes:
                options[attr] = getattr(self.titration, attr)

        options["version"] = __version__
        options["moduleNames"] = moduleNames
        options["dropdownValues"] = dropdownValues

        initialfile = os.path.splitext(self.titration.title)[0] + "_options"
        fileName = fd.asksaveasfilename(
            title="Save options to file",
            filetypes=[("NumPy archive", "*.npz")],
            initialfile=initialfile,
            defaultextension=".npz",
        )
        if fileName == "":
            return

        np.savez(fileName, **options)

    def getSelectedOptions(self, popup):
        popup.selectedOptions = []
        for name, checkbutton in popup.checkbuttons.items():
            if checkbutton.instate(["selected"]):
                popup.selectedOptions.append(name)
        popup.saved = True
        popup.destroy()

    def saveData(self):
        initialfile = os.path.splitext(self.titration.title)[0] + "_processed_input"
        fileName = fd.asksaveasfilename(
            title="Save processed input data as a universal CSV file",
            initialfile=initialfile,
            filetypes=[("CSV file", "*.csv")],
            defaultextension=".csv",
        )
        data = self.titration.processedData
        rowTitles = np.atleast_2d(self.titration.processedAdditionTitles).T
        columnTitles = np.append("", self.titration.processedSignalTitles)
        output = np.vstack((columnTitles, np.hstack((rowTitles, data))))
        np.savetxt(fileName, output, fmt="%s", delimiter=",")


class TitrationFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.numFits = 0

        # options bar on the left
        scrolledFrame = ScrolledFrame(self)
        scrolledFrame.grid(column=0, row=0, sticky="nesw")
        self.options = scrolledFrame.display_widget(
            ttk.Frame, stretch=True, padding=(0, 0, padding, 0)
        )

        self.moduleFrames = {}
        for mod in titrationModules:
            moduleFrame = mod.ModuleFrame(
                self.options, self.titration, self.updatePlots
            )
            self.moduleFrames[mod.__name__] = moduleFrame
            moduleFrame.grid(sticky="nesw", pady=padding, ipady=padding)

        fitDataButton = ttk.Button(self.options, text="Fit", command=self.tryFitData)
        fitDataButton.grid(sticky="nesw", pady=padding, ipady=padding)

        separator = ttk.Separator(self.options, orient="horizontal")
        separator.grid(sticky="nesw", pady=padding)

        saveLoadFrame = SaveLoadFrame(self.options, self.titration, self.moduleFrames)
        saveLoadFrame.grid(sticky="nesw")

        self.options.columnconfigure(0, weight=1)

        # tabs with different fits
        self.notebook = InteractiveNotebook(
            self, padding=padding, style="Flat.Interactive.TNotebook"
        )
        self.notebook.grid(column=1, row=0, sticky="nesw")

        if self.titration.continuous:
            self.inputSpectraFrame = InputSpectraFrame(self.notebook, self.titration)
            self.notebook.add(self.inputSpectraFrame, text="Input Spectra")

        self.rowconfigure(0, weight=1)
        self.columnconfigure(1, weight=1)

    def updatePlots(self):
        if hasattr(self, "inputSpectraFrame"):
            self.inputSpectraFrame.plot()

    def tryFitData(self, *args, **kwargs):
        try:
            self.fitData(*args, **kwargs)
        except Exception as e:
            mb.showerror(title="Failed to fit data", message=e, parent=self)
            return

    def fitData(self):
        self.titration.fitData()
        self.numFits += 1
        nb = ttk.Notebook(self, padding=padding, style="Flat.TNotebook")
        self.notebook.add(nb, text=f"Fit {self.numFits}")

        if self.titration.continuous:
            continuousFittedFrame = ContinuousFittedFrame(nb, self.titration)
            nb.add(continuousFittedFrame, text="Fitted Spectra")
            discreteFittedFrame = DiscreteFromContinuousFittedFrame(nb, self.titration)
            nb.add(
                discreteFittedFrame, text=f"Fit at select {self.titration.xQuantity}"
            )
        else:
            discreteFittedFrame = DiscreteFittedFrame(nb, self.titration)
            nb.add(discreteFittedFrame, text="Fitted signals")

        speciationFrame = SpeciationFrame(nb, self.titration)
        nb.add(speciationFrame, text="Speciation")

        resultsFrame = ResultsFrame(nb, self.titration)
        nb.add(resultsFrame, text="Results")

        self.notebook.select(str(nb))
        nb.select(str(discreteFittedFrame))


class InputSpectraFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration

        rangeSelection = ttk.Frame(self)
        rangeSelection.grid(row=0, column=0, sticky="")

        ttk.Label(rangeSelection, text=f"Range of {titration.xQuantity} to fit:").pack(
            side="left"
        )

        minWL = self.titration.signalTitles.min()
        maxWL = self.titration.signalTitles.max()
        decimals = self.titration.signalTitlesDecimals
        step = 1 / (10 ** decimals)

        self.fromSpinbox = ttk.Spinbox(
            rangeSelection, from_=minWL, to=maxWL, width=5, increment=step
        )
        self.fromSpinbox.set(f"{minWL:.{decimals}f}")
        self.fromSpinbox.pack(padx=padding, side="left")

        ttk.Label(rangeSelection, text="to").pack(side="left")

        self.toSpinbox = ttk.Spinbox(
            rangeSelection, from_=minWL, to=maxWL, width=5, increment=step
        )
        self.toSpinbox.set(f"{maxWL:.{decimals}f}")
        self.toSpinbox.pack(padx=padding, side="left")

        self.fig = Figure()
        self.ax = self.fig.add_subplot()

        ttk.Button(rangeSelection, text="Update", command=self.updateWLRange).pack(
            side="left"
        )

        canvas = FigureCanvasTkAgg(self.fig, master=self)
        canvas.draw()
        canvas.get_tk_widget().grid(row=1, column=0, sticky="")

        toolbar = NavigationToolbar2Tk(canvas, self, pack_toolbar=False)
        toolbar.update()
        toolbar.grid(row=2, column=0, sticky="w", padx=10 * padding)

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.grid_anchor("center")

        self.plot()

    def plot(self):
        ax = self.ax
        fig = self.fig
        titration = self.titration

        ax.cla()

        spectraColors = (
            ["black"] + ["#80808080"] * (titration.numAdditions - 2) + ["tab:red"]
        )
        colorCycler = cycler(color=spectraColors)
        ax.set_prop_cycle(colorCycler)

        ax.plot(titration.processedSignalTitles, titration.processedData.T)

        ax.set_xlabel(f"{titration.xQuantity} / {titration.xUnit}")
        ax.set_ylabel(f"{titration.yQuantity} / {titration.yUnit}")

        fig.tight_layout()
        fig.canvas.draw_idle()

    def updateWLRange(self):
        from_ = float(self.fromSpinbox.get())
        to = float(self.toSpinbox.get())
        self.titration.columnFilter = (self.titration.signalTitles >= from_) & (
            self.titration.signalTitles <= to
        )
        self.plot()


class ContinuousFittedFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.plot()

    def plot(self):
        titration = self.titration
        ks = self.titration.knownKs.copy()
        ks[np.isnan(ks)] = 10 ** titration.lastKs[: titration.kVarsCount()]
        self.fig = Figure()
        self.ax = self.fig.add_subplot()

        spectra = titration.lastFitResult
        names = titration.contributorNames()
        wavelengths = titration.processedSignalTitles
        for spectrum, name in zip(spectra, names):
            self.ax.plot(wavelengths, spectrum, label=name)

        ttk.Label(self, text=f"Fitted spectra (K = {ks[0]:.0f})", font="-size 15").grid(
            row=0, column=0, sticky=""
        )
        self.ax.set_xlabel(f"{titration.xQuantity} / {titration.xUnit}")
        self.ax.set_ylabel(f"molar {titration.yQuantity} / {titration.yUnit} M⁻¹")
        self.ax.legend()

        self.fig.tight_layout()

        canvas = FigureCanvasTkAgg(self.fig, master=self)
        canvas.draw()
        canvas.get_tk_widget().grid(row=1, column=0, sticky="")

        toolbar = NavigationToolbar2Tk(canvas, self, pack_toolbar=False)
        toolbar.update()
        toolbar.grid(row=2, column=0, sticky="w", padx=10 * padding)

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.grid_anchor("center")


class FittedFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.xQuantity = titration.freeNames[-1]
        self.xConcs = titration.lastTotalConcs.T[-1]
        self.normalisation = False
        self.logScale = False

    def populate(self):
        titration = self.titration
        self.fig = Figure()
        self.ax = self.fig.add_subplot()

        self.canvas = FigureCanvasTkAgg(self.fig, master=self)
        self.canvas.draw()
        self.canvas.get_tk_widget().grid(row=1, column=0, sticky="")

        self.toolbar = NavigationToolbar2Tk(self.canvas, self, pack_toolbar=False)
        self.toolbar.update()
        self.toolbar.grid(row=2, column=0, sticky="w", padx=10 * padding)

        ks = self.titration.knownKs.copy()
        ks[np.isnan(ks)] = 10 ** titration.lastKs[: titration.kVarsCount()]
        ttk.Label(self, text=f"Fitted curves (K = {ks[0]:.0f})", font="-size 15").grid(
            row=0, column=0, sticky=""
        )

        self.toggleButtonsFrame = ttk.Frame(self)
        self.toggleButtonsFrame.grid(row=1, column=1, sticky="")
        self.normalisationButton = ttk.Checkbutton(
            self.toggleButtonsFrame,
            text="Normalise movement",
            command=self.toggleNormalisation,
            style="Outline.Toolbutton",
        )
        self.normalisationButton.pack(pady=padding, fill="x")
        self.logScaleButton = ttk.Checkbutton(
            self.toggleButtonsFrame,
            text="Logarithmic x axis",
            command=self.toggleLogScale,
            style="Outline.Toolbutton",
        )
        self.logScaleButton.pack(pady=padding, fill="x")

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.grid_anchor("center")

    def toggleNormalisation(self):
        self.normalisation = not self.normalisation
        self.plot()

    def toggleLogScale(self):
        self.logScale = not self.logScale
        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.canvas.draw()

    def plot(self):
        self.ax.clear()

        titration = self.titration

        # xQuantity and xUnit for the fitted plot. Different from the xQuantity
        # and xUnit in the titration object, which are used for the input
        # spectra.
        xQuantity = self.xQuantity
        xUnit = titration.concsUnit
        xConcs = self.xConcs / totalConcentrations.prefixes[xUnit.strip("M")]

        if self.normalisation:
            curves = self.curves.T.copy()
            # get the largest difference from the first point for each signal
            diff = curves - curves[0]
            maxDiff = np.max(abs(diff), axis=0)
            curves = curves / maxDiff
            negatives = abs(np.amin(diff, axis=0)) > abs(np.amax(diff, axis=0))
            curves[:, negatives] *= -1
            curves = curves.T * 100

            fittedCurves = self.fittedCurves.T
            fittedCurves = fittedCurves / maxDiff
            fittedCurves[:, negatives] *= -1
            fittedCurves = fittedCurves.T * 100
            self.ax.set_ylabel(f"Normalised Δ{titration.yQuantity} / %")
        else:
            curves = self.curves
            fittedCurves = self.fittedCurves
            self.ax.set_ylabel(f"Δ{titration.yQuantity} / {titration.yUnit}")

        for curve, fittedCurve, name in zip(curves, fittedCurves, self.names):
            fittedZero = fittedCurve[0]
            curve -= fittedZero
            fittedCurve -= fittedZero
            self.ax.scatter(xConcs, curve)

            smoothX = np.linspace(xConcs.min(), xConcs.max(), 100)
            # make sure the smooth curve actually goes through all the fitted
            # points
            smoothX = np.unique(np.concatenate((smoothX, xConcs)))

            # interp1d requires all x values to be unique
            filter = np.concatenate((np.diff(xConcs).astype(bool), [True]))
            spl = interp1d(xConcs[filter], fittedCurve[filter], kind="quadratic")
            smoothY = spl(smoothX)
            self.ax.plot(smoothX, smoothY, label=name)

        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.ax.set_xlabel(f"[{xQuantity}] / {xUnit}")
        self.ax.legend()
        self.fig.tight_layout()

        self.canvas.draw()


class DiscreteFittedFrame(FittedFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.curves = self.titration.processedData.T
        self.fittedCurves = self.titration.lastFittedCurves.T
        self.names = self.titration.processedSignalTitlesStrings
        self.populate()
        self.plot()


class DiscreteFromContinuousFittedFrame(FittedFrame):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        peakIndices = self.titration.getPeakIndices()
        self.curves = self.titration.processedData.T[peakIndices]
        self.fittedCurves = self.titration.lastFittedCurves.T[peakIndices]
        peakTitles = self.titration.processedSignalTitlesStrings[peakIndices]
        self.names = [f"{title} {self.titration.xUnit}" for title in peakTitles]
        self.populate()
        self.plot()


class SpeciationFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.xQuantity = titration.freeNames[-1]
        self.xConcs = titration.lastTotalConcs.T[-1]
        self.speciesVar = tk.StringVar(self)
        self.logScale = False
        self.populate()
        self.plot()

    def populate(self):
        titration = self.titration
        self.fig = Figure()
        self.ax = self.fig.add_subplot()

        self.canvas = FigureCanvasTkAgg(self.fig, master=self)
        self.canvas.draw()
        self.canvas.get_tk_widget().grid(row=1, column=0, sticky="")

        self.toolbar = NavigationToolbar2Tk(self.canvas, self, pack_toolbar=False)
        self.toolbar.update()
        self.toolbar.grid(row=2, column=0, sticky="w", padx=10 * padding)

        self.optionsFrame = ttk.Frame(self)
        self.optionsFrame.grid(row=1, column=1, sticky="")
        self.speciesLabel = ttk.Label(
            self.optionsFrame, anchor="center", justify="center", text="Select species:"
        )
        self.speciesLabel.pack(pady=padding, fill="x")

        self.speciesDropdown = ttk.OptionMenu(
            self.optionsFrame,
            self.speciesVar,
            titration.freeNames[0],
            command=lambda *args: self.plot(),
            *titration.freeNames,
            style="primary.Outline.TMenubutton",
        )
        self.speciesDropdown.pack()
        self.logScaleButton = ttk.Checkbutton(
            self.optionsFrame,
            text="Logarithmic x axis",
            command=self.toggleLogScale,
            style="Outline.Toolbutton",
        )
        self.logScaleButton.pack(pady=padding, fill="x")

        self.rowconfigure(0, weight=1)
        self.rowconfigure(1, weight=1)
        self.rowconfigure(2, weight=1)
        self.grid_anchor("center")

    def toggleLogScale(self):
        self.logScale = not self.logScale
        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.canvas.draw()

    @property
    def freeIndex(self):
        return np.where(self.titration.freeNames == self.speciesVar.get())[0][0]

    def plot(self):
        self.ax.clear()

        titration = self.titration

        # xQuantity and xUnit for the fitted plot. Different from the xQuantity
        # and xUnit in the titration object, which are used for the input
        # spectra.
        totalConcs = titration.lastTotalConcs[:, self.freeIndex]
        additionsFilter = totalConcs != 0
        totalConcs = totalConcs[additionsFilter]

        xQuantity = self.xQuantity
        xUnit = titration.concsUnit
        xConcs = (
            self.xConcs[additionsFilter]
            / totalConcentrations.prefixes[xUnit.strip("M")]
        )

        freeConcs = titration.lastFreeConcs[additionsFilter, :][:, self.freeIndex]
        freeName = self.speciesVar.get()

        factor = abs(titration.stoichiometries[:, self.freeIndex])
        boundFilter = factor.astype(bool)

        boundConcs = titration.lastBoundConcs * factor
        boundConcs = boundConcs[additionsFilter, :][:, boundFilter]
        boundNames = titration.boundNames[boundFilter]

        curves = 100 * np.vstack((freeConcs, boundConcs.T)) / totalConcs
        names = np.append(freeName, boundNames)
        self.ax.set_ylabel(f"% of {freeName}")

        for curve, name in zip(curves, names):
            smoothX = np.linspace(xConcs.min(), xConcs.max(), 100)
            # make sure the smooth curve actually goes through all the fitted
            # points
            smoothX = np.unique(np.concatenate((smoothX, xConcs)))

            # interp1d requires all x values to be unique
            filter = np.concatenate((np.diff(xConcs).astype(bool), [True]))
            spl = interp1d(xConcs[filter], curve[filter], kind="quadratic")
            smoothY = spl(smoothX)
            self.ax.plot(smoothX, smoothY, label=name)

        if self.logScale:
            self.ax.set_xscale("log")
        else:
            self.ax.set_xscale("linear")
        self.ax.set_xlabel(f"[{xQuantity}] / {xUnit}")
        self.ax.legend()
        self.fig.tight_layout()

        self.canvas.draw()


class ResultsFrame(ttk.Frame):
    def __init__(self, parent, titration, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)
        self.titration = titration
        self.showResults()

    def showResults(self):
        titration = self.titration
        kTable = Table(
            self,
            0,
            0,
            ["K (M⁻ⁿ)", "α"],
            rowOptions=("readonlyTitles",),
            columnOptions=("readonlyTitles",),
        )

        ks = self.titration.knownKs.copy()
        ks[np.isnan(ks)] = 10 ** titration.lastKs[: titration.kVarsCount()]

        alphas = titration.knownAlphas.copy()
        polymerAlphas = alphas[np.any(titration.stoichiometries < 0, 1)]
        polymerAlphas[np.isnan(polymerAlphas)] = (
            10
            ** titration.lastKs[titration.kVarsCount() + titration.getConcVarsCount() :]
        )

        for boundName, k, alpha in zip(self.titration.boundNames, ks, alphas):
            kTable.addRow(boundName, [np.rint(k), alpha if not np.isnan(alpha) else ""])
        kTable.pack(side="top", pady=15)

        sheet = tksheet.Sheet(
            self,
            data=list(np.around(titration.lastFitResult, 2)),
            headers=list(titration.processedSignalTitlesStrings),
            row_index=list(titration.contributorNames()),
            set_all_heights_and_widths=True,
        )
        sheet.enable_bindings()
        sheet.pack(side="top", pady=15, fill="x")

        saveButton = ttk.Button(
            self, text="Save as CSV", command=self.saveCSV, style="success.TButton"
        )
        saveButton.pack(side="top", pady=15)

    def saveCSV(self):
        initialfile = os.path.splitext(self.titration.title)[0] + "_fit"
        fileName = fd.asksaveasfilename(
            title="Save fitted spectra",
            initialfile=initialfile,
            filetypes=[("CSV file", "*.csv")],
            defaultextension=".csv",
        )
        data = self.titration.lastFitResult
        rowTitles = np.atleast_2d(self.titration.contributorNames()).T
        columnTitles = np.append("", self.titration.processedSignalTitles)
        output = np.vstack((columnTitles, np.hstack((rowTitles, data))))
        np.savetxt(fileName, output, fmt="%s", delimiter=",")



import numpy as np
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.messagebox as mb
from tkinter import font

from . import moduleFrame
from .table import Table, ButtonFrame
from .scrolledFrame import ScrolledFrame


class CustomKsTable(Table):
    def __init__(self, master, titration):
        self.titration = titration
        if hasattr(titration, "kVarsNames"):
            kVarsNames = titration.kVarsNames
        else:
            kVarsNames = [f"K_{name}" for name in titration.boundNames]
        if hasattr(titration, "ksMatrix"):
            ksMatrix = titration.ksMatrix
        else:
            ksMatrix = np.identity(titration.boundCount, dtype=int)
        if (
            hasattr(titration, "knownKs")
            and len(titration.knownKs) == ksMatrix.shape[0]
        ):
            knownKs = np.where(np.isnan(titration.knownKs), "", titration.knownKs)
        else:
            knownKs = np.full(ksMatrix.shape[0], "")
        columnTitles = np.append(
            [f"Global K for {name}" for name in titration.boundNames], "Value"
        )
        self.width = max([len(title) for title in columnTitles]) + 1

        super().__init__(
            master,
            0,
            0,
            columnTitles,
            allowBlanks=True,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
            callback=self.createLabels,
        )
        self.addConstantsRow()
        for name, contributions, value in zip(kVarsNames, ksMatrix, knownKs):
            self.addRow(name, np.append(contributions, value))

    def addConstantsRow(self):
        if hasattr(self.titration, "ksConstants"):
            ksConstants = self.titration.ksConstants
        else:
            ksConstants = np.append(np.full(len(self.columnTitles) - 1, "1"), "")
        oldRowOptions = self.rowOptions
        self.rowOptions = ("readonlyTitles",)
        self.addRow("Statistical factor", ksConstants)
        self.rowOptions = oldRowOptions

    def createLabels(self, *args, **kwargs):
        try:
            labels = []
            superscripts = "⁰¹²³⁴⁵⁶⁷⁸⁹"
            variables = self.rowTitles[1:]
            for globalK, statFactor, variableFactors in zip(
                self.columnTitles[:-1],
                self.data[0, :-1],
                self.data[1:, :-1].T.astype(int),
            ):
                label = f"{globalK} = {statFactor:g}"
                for variable, factor in zip(variables, variableFactors):
                    if factor == 0 or factor == "":
                        continue
                    label += f" × {variable}"
                    if factor == 1:
                        continue
                    label += "".join(
                        [superscripts[int(digit)] for digit in str(factor)]
                    )
                labels.append(label)
            self.label.configure(text="\n".join(labels))
        except Exception:
            pass
        return True

    def convertData(self, number):
        if number == "":
            return np.nan
        else:
            return float(number)


class CustomKsPopup(tk.Toplevel):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter relationships between Ks")

        height = int(self.master.winfo_height() * 0.4)
        frame = ScrolledFrame(self, height=height, max_width=1500)
        frame.pack(expand=True, fill="both")

        self.innerFrame = frame.display_widget(ttk.Frame, stretch=True)

        self.customKsTable = CustomKsTable(self.innerFrame, titration)
        self.customKsTable.pack(expand=True, fill="both")

        self.labelFont = font.nametofont("TkTextFont").copy()
        self.labelFont["size"] = int(1.3 * self.labelFont["size"])

        self.equationsLabel = ttk.Label(
            self.innerFrame, anchor="center", font=self.labelFont, padding=5
        )
        self.equationsLabel.pack(expand=True, fill="both")
        self.customKsTable.label = self.equationsLabel
        self.customKsTable.createLabels()

        buttonFrame = ButtonFrame(
            self.innerFrame, self.reset, self.saveData, self.destroy
        )
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        self.customKsTable.destroy()
        self.customKsTable = CustomKsTable(self.innerFrame, self.titration)
        self.customKsTable.pack(before=self.equationsLabel, expand=True, fill="both")
        self.customKsTable.label = self.equationsLabel
        self.customKsTable.createLabels()

    def saveData(self):
        try:
            _contributorsMatrix = self.customKsTable.data
        except Exception as e:
            mb.showerror(title="Could not save data", message=e, parent=self)
            return

        self.titration._contributorsMatrix = _contributorsMatrix
        self.titration._contributorNames = self.customKsTable.rowTitles

        self.destroy()


class GetKsCustom(moduleFrame.Strategy):
    popup = CustomKsPopup

    popupAttributes = ("ksMatrix", "knownKs")

    def __init__(self, titration):
        self.titration = titration
        if not (
            hasattr(titration, "knownKs")
            and len(titration.knownKs) == titration.boundCount
        ):
            titration.knownKs = np.full(titration.boundCount, np.nan)
        if not (
            hasattr(titration, "knownAlphas")
            and len(titration.knownAlphas) == titration.boundCount
        ):
            titration.knownAlphas = np.full(titration.boundCount, np.nan)
        titration.kVarsCount = self.kVarsCount
        titration.alphaVarsCount = self.alphaVarsCount

    def __call__(self, kVars, alphaVars):
        ks = self.titration.knownKs.copy()
        ks[np.isnan(ks)] = kVars
        alphas = self.titration.knownAlphas[self.titration.polymerIndices].copy()
        alphas[np.isnan(alphas)] = alphaVars
        return (ks, alphas)

    def kVarsCount(self):
        return np.count_nonzero(np.isnan(self.titration.knownKs))

    def alphaVarsCount(self):
        return np.count_nonzero(
            np.isnan(self.titration.knownAlphas[self.titration.polymerIndices])
        )


class KnownKsTable(Table):
    def __init__(self, master, titration):
        self.titration = titration
        super().__init__(
            master,
            0,
            0,
            ["Value", "α"],
            allowBlanks=True,
            rowOptions=("readonlyTitles",),
            columnOptions=("readonlyTitles",),
        )
        self.populateDefault()

    def populateDefault(self):
        for boundName, knownK, knownAlpha in zip(
            self.titration.boundNames,
            self.titration.knownKs,
            self.titration.knownAlphas,
        ):
            self.addRow(
                boundName,
                [
                    knownK if not np.isnan(knownK) else "",
                    knownAlpha if not np.isnan(knownAlpha) else "",
                ],
            )


class KnownKsPopup(tk.Toplevel):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter known equilibrium constants")

        height = int(self.master.winfo_height() * 0.4)
        frame = ScrolledFrame(self, height=height, max_width=1500)
        frame.pack(expand=True, fill="both")

        innerFrame = frame.display_widget(ttk.Frame, stretch=True)
        knownKsLabel = ttk.Label(
            innerFrame, text="Enter known K values. Leave blank for unknown values."
        )
        knownKsLabel.pack()

        self.knownKsTable = KnownKsTable(innerFrame, titration)
        self.knownKsTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(innerFrame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

    def reset(self):
        self.knownKsTable.resetData()
        self.knownKsTable.columnTitles = ["Value", "α"]
        self.knownKsTable.populateDefault()

    def saveData(self):
        try:
            knownKs = self.knownKsTable.data[:, 0]
            knownAlphas = self.knownKsTable.data[:, 1]
        except Exception as e:
            mb.showerror(title="Could not save data", message=e, parent=self)
            return

        self.titration.knownKs = knownKs
        self.titration.knownAlphas = knownAlphas
        self.destroy()


class GetKsKnown(moduleFrame.Strategy):
    popup = KnownKsPopup
    popupAttributes = ("knownKs", "knownAlphas")

    def __init__(self, titration):
        self.titration = titration
        if not (
            hasattr(titration, "knownKs")
            and len(titration.knownKs) == titration.boundCount
        ):
            titration.knownKs = np.full(titration.boundCount, np.nan)
        if not (
            hasattr(titration, "knownAlphas")
            and len(titration.knownAlphas) == titration.boundCount
        ):
            titration.knownAlphas = np.full(titration.boundCount, np.nan)
        titration.kVarsCount = self.kVarsCount
        titration.alphaVarsCount = self.alphaVarsCount

    def __call__(self, kVars, alphaVars):
        ks = self.titration.knownKs.copy()
        ks[np.isnan(ks)] = kVars
        alphas = self.titration.knownAlphas[self.titration.polymerIndices].copy()
        alphas[np.isnan(alphas)] = alphaVars
        return (ks, alphas)

    def kVarsCount(self):
        return np.count_nonzero(np.isnan(self.titration.knownKs))

    def alphaVarsCount(self):
        return np.count_nonzero(
            np.isnan(self.titration.knownAlphas[self.titration.polymerIndices])
        )


class GetKsNonspecific(moduleFrame.Strategy):
    def __init__(self, titration):
        self.titration = titration
        titration.kVarsCount = self.kVarsCount
        titration.alphaVarsCount = self.alphaVarsCount

    def __call__(self, kVars, alphaVars):
        # TODO: consider replacing fixed value with percentage of 2:1 K
        arbitrarilySmallK = 0.001
        self.titration.knownKs = np.full(self.titration.boundCount, np.nan)
        self.titration.knownKs[self.trimerIndices()] = arbitrarilySmallK
        self.titration.knownAlphas = np.full(self.titration.boundCount, np.nan)
        ks = np.full(self.titration.boundCount, arbitrarilySmallK)
        ks[~self.trimerIndices()] = kVars
        return (ks, alphaVars)

    def trimerIndices(self):
        return self.titration.stoichiometries.sum(axis=1) == 3

    def kVarsCount(self):
        return self.titration.boundCount - np.count_nonzero(self.trimerIndices())

    def alphaVarsCount(self):
        return self.titration.polymerCount


class GetKsAll(moduleFrame.Strategy):
    # when every equilibrium constant is unknown and independent
    def __init__(self, titration):
        self.titration = titration
        titration.kVarsCount = self.kVarsCount
        titration.alphaVarsCount = self.alphaVarsCount

    def alphaVarsCount(self):
        return self.titration.polymerCount

    def kVarsCount(self):
        return self.titration.boundCount

    def __call__(self, kVars, alphaVars):
        # TODO: move this to a more sensible place
        self.titration.knownKs = np.full(self.titration.boundCount, np.nan)
        self.titration.knownAlphas = np.full(self.titration.boundCount, np.nan)
        return (kVars, alphaVars)


class ModuleFrame(moduleFrame.ModuleFrame):
    frameLabel = "Equilibrium constants"
    dropdownLabelText = "Fix any K values?"
    dropdownOptions = {
        "No, optimise all Ks": GetKsAll,
        "Assume second binding weak": GetKsNonspecific,
        "Fix some known Ks": GetKsKnown,
        "Custom": GetKsCustom,
    }
    attributeName = "getKs"



from . import moduleFrame


class GetConcs(moduleFrame.Strategy):
    def __call__(self, signalVars, totalConcs):
        return signalVars


class GetFraction(moduleFrame.Strategy):
    def __call__(self, signalVars, totalConcs):
        # divide by total host concs
        # TODO: support proportionality to other total concs
        return signalVars / totalConcs[:, [0]]


class ModuleFrame(moduleFrame.ModuleFrame):
    frameLabel = "Proportionality"
    dropdownLabelText = "What are the signals proportional to?"
    dropdownOptions = {
        "Concentration (slow exchange)": GetConcs,
        "Mole fraction (fast exchange)": GetFraction,
    }
    attributeName = "getProportionalSignals"
    setDefault = False



import numpy as np
from numpy.linalg import lstsq
from scipy.optimize import lsq_linear

from . import moduleFrame


class FitSignals(moduleFrame.Strategy):
    def __call__(self, signalVars, knownSpectra):
        # rows are additions, columns are contributors
        knownMask = ~np.isnan(knownSpectra[:, 0])
        knownSignals = signalVars[:, knownMask]
        unknownSignals = signalVars[:, ~knownMask]

        knownSpectrum = knownSignals @ knownSpectra[knownMask, :]
        unknownSpectrum = self.titration.processedData - knownSpectrum
        fittedSignals, residuals, _, _ = lstsq(
            unknownSignals, unknownSpectrum, rcond=None
        )
        fittedCurves = unknownSignals @ fittedSignals + knownSpectrum
        allSignals = knownSpectra.copy()
        allSignals[~knownMask, :] = fittedSignals
        return allSignals, residuals, fittedCurves


class FitSignalsNonnegative(moduleFrame.Strategy):
    # TODO: account for known spectra
    def __call__(self, signalVars, knownSpectra):
        fittedSignals = np.empty((0, signalVars.shape[1]))
        residuals = np.empty((1, 0))
        for signal in self.titration.processedData.T:
            result = lsq_linear(signalVars, signal, (0, np.inf), method="bvls")
            fittedSignals = np.vstack((fittedSignals, result.x))
            residuals = np.append(residuals, result.cost)
        fittedSignals = fittedSignals.T
        fittedCurves = signalVars @ fittedSignals
        return fittedSignals, residuals, fittedCurves


class ModuleFrame(moduleFrame.ModuleFrame):
    frameLabel = "Fit signals"
    dropdownLabelText = "Fit signals to curve using:"
    # TODO: add least squares with linear constraints
    dropdownOptions = {
        "Ordinary least squares": FitSignals,
        "Nonnegative least squares": FitSignalsNonnegative,
    }
    attributeName = "fitSignals"



import tkinter as tk
import tkinter.ttk as ttk


# all module strategies should be a subclass
class Strategy:
    popup = None
    # List of titration attributes that are set through the popup window, and
    # can be loaded from / saved to a file.
    popupAttributes = ()

    def __init__(self, titration):
        self.titration = titration

    def __call__(self, *args, **kwargs):
        raise NotImplementedError()


class ModuleFrame(ttk.LabelFrame):
    frameLabel = ""
    dropdownLabelText = ""
    dropdownOptions = {}
    attributeName = ""
    setDefault = True

    def __init__(self, parent, titration, updatePlots, *args, **kwargs):
        super().__init__(parent, text=self.frameLabel, borderwidth=5, *args, **kwargs)
        self.titration = titration
        self.updatePlots = updatePlots

        self.stringVar = tk.StringVar()

        self.dropdownLabel = ttk.Label(self, text=self.dropdownLabelText)
        self.dropdownLabel.pack()

        strategies = list(self.dropdownOptions.keys())
        defaultValue = strategies[0] if self.setDefault else None
        optionMenu = ttk.OptionMenu(
            self,
            self.stringVar,
            defaultValue,
            command=self.callback,
            *strategies,
            style="primary.Outline.TMenubutton"
        )
        optionMenu.configure(width=30)
        # call the callback for the default value
        if self.setDefault:
            self.callback(defaultValue)
        optionMenu.pack(fill="x", pady=(5, 0))

    def callback(self, value):
        SelectedStrategy = self.dropdownOptions[value]
        selectedStrategy = SelectedStrategy(self.titration)
        if selectedStrategy.popup is not None:
            popup = selectedStrategy.popup(self.titration)
            # On some versions of macOS, calling grab_set here sets the popup
            # grab before the global grab on the menu is realeased, leaving
            # the popup unresponsive to mouse events.
            # TODO: find better place for callback that avoids this issue
            if self._windowingsystem != "aqua":
                popup.grab_set()
            popup.wait_window()
        setattr(self.titration, self.attributeName, selectedStrategy)
        self.updatePlots()



import numpy as np
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.messagebox as mb

from . import moduleFrame
from .table import Table, ButtonFrame
from .scrolledFrame import ScrolledFrame
from .style import padding


class GetSignalVarsFromMatrix(moduleFrame.Strategy):
    def __init__(self, titration):
        self.titration = titration
        self.titration.contributorsMatrix = self.contributorsMatrix
        self.titration.contributorNames = self.contributorNames

    def __call__(self, freeConcs, boundConcs):
        allConcs = np.concatenate((freeConcs, boundConcs), 1)
        variableConcs = allConcs @ self.titration.contributorsMatrix().T
        return variableConcs


class ContributorsTable(Table):
    def __init__(self, master, titration):
        columnTitles = np.concatenate((titration.freeNames, titration.boundNames))
        self.width = max([len(title) for title in columnTitles]) + 5
        super().__init__(
            master,
            0,
            0,
            columnTitles,
            allowBlanks=False,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
        )
        for name, contributions in zip(
            titration.contributorNames(), titration.contributorsMatrix()
        ):
            self.addRow(name, contributions)

    def convertData(self, number):
        if number.isdigit():
            return int(number)
        else:
            return float(number)


class ContributorsPopup(tk.Toplevel):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter the contributing states")

        height = int(self.master.winfo_height() * 0.4)
        frame = ScrolledFrame(self, height=height, max_width=1500)
        frame.pack(expand=True, fill="both")

        self.innerFrame = frame.display_widget(ttk.Frame, stretch=True)

        self.contributorsTable = ContributorsTable(self.innerFrame, titration)
        self.contributorsTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(
            self.innerFrame, self.reset, self.saveData, self.destroy
        )
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        self.update()
        self.contributorsLabel = ttk.Label(
            self.innerFrame,
            wraplength=self.innerFrame.winfo_reqwidth(),
            padding=padding * 2,
            text=(
                "On each row, enter a state that contributes to the observed"
                " signal. For each column, specify how many of the state that"
                " species contains."
            ),
        )
        self.contributorsLabel.pack(
            before=self.contributorsTable, expand=True, fill="both"
        )

    def reset(self):
        self.contributorsTable.destroy()
        self.contributorsTable = ContributorsTable(self.innerFrame, self.titration)
        self.contributorsTable.pack(expand=True, fill="both")

    def saveData(self):
        try:
            _contributorsMatrix = self.contributorsTable.data
        except Exception as e:
            mb.showerror(title="Could not save data", message=e, parent=self)
            return

        self.titration._contributorsMatrix = _contributorsMatrix
        self.titration._contributorNames = self.contributorsTable.rowTitles

        self.destroy()


class GetSignalVarsCustom(GetSignalVarsFromMatrix):
    popup = ContributorsPopup
    popupAttributes = ("_contributorsMatrix", "_contributorNames")

    def __init__(self, titration, *args, **kwargs):
        titration._contributorsMatrix = titration.contributorsMatrix()
        titration._contributorNames = titration.contributorNames()
        super().__init__(titration, *args, **kwargs)

    def contributorsMatrix(self):
        return self.titration._contributorsMatrix

    def contributorNames(self):
        return self.titration._contributorNames


class GetSignalVarsAll(moduleFrame.Strategy):
    def __init__(self, titration):
        self.titration = titration
        self.titration.contributorsMatrix = self.contributorsMatrix
        self.titration.contributorNames = self.contributorNames

    def contributorNames(self):
        return np.concatenate((self.titration.freeNames, self.titration.boundNames))

    def contributorsMatrix(self):
        totalCount = self.titration.freeCount + self.titration.boundCount
        return np.identity(totalCount, dtype=int)

    def __call__(self, freeConcs, boundConcs):
        allConcs = np.concatenate((freeConcs, boundConcs), 1)
        return allConcs


class GetSignalVarsSingle(GetSignalVarsFromMatrix):
    def filter(self):
        hostFree = np.zeros(self.titration.freeCount, dtype=int)
        hostFree[self.contributorIndex] = 1

        hostBound = abs(self.titration.stoichiometries[:, self.contributorIndex])

        return np.concatenate((hostFree, hostBound))

    def filter_binary(self):
        return self.filter().astype(bool).astype(int)

    def contributorNames(self):
        allNames = np.concatenate((self.titration.freeNames, self.titration.boundNames))
        return allNames[self.filter().astype(bool)]

    def contributorsMatrix(self):
        matrix = np.diag(self.filter())
        emptyRows = np.all(matrix == 0, axis=1)
        return matrix[~emptyRows]


class GetSignalVarsHost(GetSignalVarsSingle):
    contributorIndex = 0


class ModuleFrame(moduleFrame.ModuleFrame):
    frameLabel = "Contributors"
    dropdownLabelText = "What contributes to the signals?"
    dropdownOptions = {
        "Only Host": GetSignalVarsHost,
        "All": GetSignalVarsAll,
        "Custom": GetSignalVarsCustom,
    }
    attributeName = "getSignalVars"



import numpy as np
import tkinter as tk
import tkinter.ttk as ttk
import tkinter.messagebox as mb

from tksheet import Sheet

from . import moduleFrame
from .table import ButtonFrame
from .style import padding


class KnownSpectraPopup(tk.Toplevel):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter known spectra")

        self.sheet = Sheet(
            self,
            data=self.formatData(titration.knownSpectra),
            headers=list(titration.processedSignalTitlesStrings),
            row_index=list(titration.contributorNames()),
            set_all_heights_and_widths=True,
        )

        self.sheet.pack(side="top", expand=True, fill="both")
        self.sheet.enable_bindings()

        buttonFrame = ButtonFrame(self, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        loadButton = ttk.Button(buttonFrame, text="Load from CSV", command=self.loadCSV)
        loadButton.pack(side="left", padx=padding)

    def formatData(self, data):
        formattedData = data.astype(str)
        formattedData[np.isnan(data)] = ""
        return list(formattedData)

    def reset(self):
        self.sheet.set_sheet_data(self.formatData(self.titration.knownSpectra))

    def saveData(self):
        try:
            data = np.array(self.sheet.get_sheet_data(), dtype=object)
            data[np.where(data == "")] = None
            data = data.astype(float)
        except Exception as e:
            mb.showerror(title="Could not save data", message=e, parent=self)
            return

        if not np.all(np.any(np.isnan(data), 1) == np.all(np.isnan(data), 1)):
            mb.showerror(
                title="Could not save data",
                parent=self,
                message="Please enter full spectra, or leave the entire row blank",
            )
            return

        self.titration.knownSpectra = data
        self.destroy()

    def loadCSV(self):
        return


class GetKnownSpectra(moduleFrame.Strategy):
    popup = KnownSpectraPopup
    popupAttributes = ("knownSpectra",)

    def __init__(self, titration):
        self.titration = titration
        if not (
            hasattr(titration, "knownSpectra")
            and titration.knownSpectra.shape
            == (len(titration.contributorNames()), len(titration.processedSignalTitles))
        ):
            titration.knownSpectra = np.full(
                (
                    len(self.titration.contributorNames()),
                    len(self.titration.processedSignalTitles),
                ),
                np.nan,
            )

    def __call__(self):
        return self.titration.knownSpectra


class GetAllSpectra(moduleFrame.Strategy):
    def __call__(self):
        return np.full(
            (
                len(self.titration.contributorNames()),
                len(self.titration.processedSignalTitles),
            ),
            np.nan,
        )


class ModuleFrame(moduleFrame.ModuleFrame):
    frameLabel = "Spectra"
    dropdownLabelText = "Which spectra to optimise?"
    dropdownOptions = {
        "Optimise all spectra": GetAllSpectra,
        "Specify some known spectra": GetKnownSpectra,
    }
    attributeName = "getKnownSpectra"
